/*******************************************************************************
 * Copyright 2024 ModalAI Inc.
 ******************************************************************************/

/**
 * @file       mcv-test-pyr-fpx.cpp
 *
 * @brief      Tests the PYR Feature Extraction (FPX) process
 *
 * @author     Thomas Patton (thomas.patton@modalai.com)
 * @date       2024
 */

#include <fastcv.h>
#include <fcntl.h>
#include <getopt.h>
#include <modalcv.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>   // for mkdir
#include <sys/types.h>  // for mode_t in mkdir
#include <time.h>
#include <unistd.h>

#include <iomanip>
#include <iostream>
#include <opencv2/imgcodecs.hpp>
#include <opencv2/imgproc.hpp>
#include <sstream>

#include "modalcv/io.hpp"
#include "modalcv/pyr.h"

// output files
#define OUTPUT_FOLDER "/data/libmodalcv/pyr_fpx/"

/**
 * @brief Helper fn for getting the current clock time in nanoseconds
 *
 * @return Time in ns
 */
static int64_t _apps_time_monotonic_ns() {
    struct timespec ts;
    if (clock_gettime(CLOCK_MONOTONIC, &ts)) {
        fprintf(stderr, "ERROR calling clock_gettime\n");
        return -1;
    }
    return (int64_t)ts.tv_sec * 1000000000 + (int64_t)ts.tv_nsec;
}

unsigned char* getAlignedImageData(const cv::Mat& image, bool& needToFree);
uint32_t* allocateAlignedUint32Array(size_t N);
void freeAlignedMemory(void* ptr);

unsigned char* getAlignedData(const cv::Mat& image, bool& needToFree) {
    if (reinterpret_cast<uintptr_t>(image.data) % 16 == 0) {
        needToFree = false;
        return image.data;
    } else {
        size_t totalSize = image.total() * image.elemSize();
        unsigned char* alignedData = nullptr;
        if (posix_memalign(reinterpret_cast<void**>(&alignedData), 16, totalSize) != 0) {
            needToFree = false;
            return nullptr; // Allocation failed
        }
        memcpy(alignedData, image.data, totalSize);
        needToFree = true;
        return alignedData;
    }
}


int main(int argc, char* argv[]) {
    mcv_en_debug_prints(0);
    mcv_en_timing_prints(0);

    int w, h;
    std::vector<std::string> paths;
    if (init_modalcv_example(argc, argv, paths, w, h)) {
        fprintf(stderr, "ERROR, failed to initialize example\n");
        return -1;
    }

    std::string input_folder = argv[1];
    printf("%s\n", input_folder.c_str());

    // configure everything
    mcv_cvp_pyr_fpx_config_t pyr_fpx_config = mcv_cvp_pyr_fpx_default_config();
    pyr_fpx_config.width = w;
    pyr_fpx_config.height = h;

    // create a pyr_fpx handle from our configuration
    mcv_cvp_pyr_fpx_handle handle = mcv_cvp_pyr_fpx_init(pyr_fpx_config);
    if (NULL == handle) {
        fprintf(stderr, "ERROR creating pyr fpx handle\n");
    }

    // create an output object
    mcv_cvp_pyr_fpx_t output_pyr_fpx;
    if (mcv_cvp_create_pyr_fpx(handle, &output_pyr_fpx)) {
        fprintf(stderr, "ERROR creating feature extraction pyramid\n");
        return -1;
    }

    // MAIN LOOP iterate over all images in folder
    printf("STARTING MAIN LOOP\n");
    for (size_t i = 0; i < paths.size(); i++) {
        std::stringstream ss;
        ss << std::setw(5) << std::setfill('0') << i;
        std::string input_path = paths[i];
        printf("%s\n", input_path.c_str());

        cv::Mat image, current_img_color;
        image = imread(input_path, cv::IMREAD_GRAYSCALE);
        
        // cv::GaussianBlur(image, image, cv::Size(3, 3), 0);
        // cv::equalizeHist(image, image);
        cv::Mat filtered_img;
        // cv::bilateralFilter(image, filtered_img, 9, 75, 75);
        // image = filtered_img.clone();

        cv::cvtColor(image, current_img_color, cv::COLOR_GRAY2BGR);
        if (image.type() != CV_8UC1) {
            fprintf(stderr, "ERROR loading images\n");
            return -1;
        }

        // run the pyr fpx process on the current frame
        int64_t start = _apps_time_monotonic_ns();
        mcv_cvp_pyr_fpx_process(handle, image.data, &output_pyr_fpx);

        // fast cv feature extraction options
        // fcvCornerFast9u8()
        // fcvCornerHarrisu8()

        int width = image.cols;
        int height = image.rows;

        bool needToFreeSrcData = false;
        unsigned char* srcData = getAlignedData(image, needToFreeSrcData);
        if (srcData == nullptr) {
            std::cerr << "Failed to allocate aligned memory for srcData.\n";
            return -1;
        }

        // Prepare output array
        unsigned int nCornersMax = 100000;
        size_t xySize = 2 * nCornersMax * sizeof(uint32_t);
        uint32_t* xy = nullptr;
        if (posix_memalign(reinterpret_cast<void**>(&xy), 16, xySize) != 0) {
            std::cerr << "Failed to allocate aligned memory for xy.\n";
            if (needToFreeSrcData) free(srcData);
            return -1;
        }

        float* harrisResp = (float*)malloc(width*height*sizeof(float));

        size_t stride = image.step;
        if (stride % 8 != 0) {
            cv::Mat tempImage(height, width, CV_8UC1);
            image.copyTo(tempImage);
            image = tempImage;
            stride = image.step;
        }

        unsigned int nCornersDetected = 0;

        // fcvCornerFast9u8(srcData, 
        //                  width,
        //                  height,
        //                  stride,
        //                  20,
        //                  3,
        //                  xy,
        //                  nCornersMax,
        //                  &nCornersDetected);

        // fcvGoodFeatureToTracku8(
        //     srcData,
        //     width,
        //     height,
        //     stride,
        //     3.f,
        //     20,
        //     50,
        //     xy,
        //     nCornersMax,
        //     &nCornersDetected
        // );

        // fcvCornerHarrisu8
        // fcvCornerHarrisAdaptiveu8
        // has thresholding, still gives too many features
        // fcvCornerHarrisu8(srcData,
        //                     width,
        //                     height,
        //                     stride,
        //                     20, // border
        //                     xy,
        //                     nCornersMax,
        //                     &nCornersDetected,
        //                     1000); // threshold

        // fcvCornerHarrisScoreu8(srcData,
        //                        width,
        //                        height,
        //                        stride,
        //                        harrisResp,
        //                        0,
        //                        xy,
        //                        nCornersMax,
        //                        &nCornersDetected,
        //                        100, // normalized threshold
        //                        0.04,
        // printf("Iter Time ::    %6.2fms\n", (double)(end - start) / 1000000);
        //                        5, // kernel size for gradient computation
        //                        2, // block size??
        //                        0,3,
        //                        1);

        if (needToFreeSrcData) {
            free(srcData);
        }
        int64_t end = _apps_time_monotonic_ns();
        printf("Iter Time ::    %6.2fms\n", (double)(end - start) / 1000000);

        for (unsigned int j = 0; j < nCornersDetected; ++j) {
            uint32_t x = xy[2 * j];
            uint32_t y = xy[2 * j + 1];
            // printf("i: %4d, (%4d, %4d)\n", j, x, y);
            cv::circle(current_img_color, cv::Point(x, y), 3, cv::Scalar(255), -1);
        }

        printf("number of corners: %d\n", nCornersDetected);
        free(xy);
        free(harrisResp);

        
        int points_found = 0;
        // draw features on the image
        for (size_t oct = 0; oct < pyr_fpx_config.n_octaves; oct++) {
            for (int i = 0; i < output_pyr_fpx.n_points[oct]; i++) {
                points_found++;

                mcv_cvp_fpx_feature_t feature = output_pyr_fpx.fpx[oct][i];
                int feature_x = feature.x * pow(2, oct);  // upscale
                int feature_y = feature.y * pow(2, oct);  // upscale
                int radius = 4;
                int thickness = 1;

                cv::Scalar color;
                if (oct == 0) {
                    color = cv::Scalar(0, 255, 0);
                } else if (oct == 1) {
                    color = cv::Scalar(255, 0, 0);
                } else {
                    color = cv::Scalar(0, 0, 255);
                }

                cv::Rect rect(feature_x - radius, feature_y - radius,
                              radius * 2, radius * 2);
                cv::rectangle(current_img_color, rect, color, thickness, 1, 0);
            }
        }


        std::string output_path = OUTPUT_FOLDER + ss.str() + ".png";
        cv::imwrite(output_path, current_img_color);
    }

    // cleanup
    printf("all done, deinitializing\n");
    mcv_cvp_destroy_pyr_fpx(handle, &output_pyr_fpx);
    mcv_cvp_pyr_fpx_deinit(handle);
    printf("DONE\n");
    return 0;
}

unsigned char* getAlignedImageData(const cv::Mat& image, bool& needToFree) {
    // Check if the data is already 16-byte aligned
    if (reinterpret_cast<uintptr_t>(image.data) % 16 == 0) {
        // Data is already 16-byte aligned
        needToFree = false;
        return image.data;
    } else {
        // Allocate a new buffer with 16-byte alignment
        size_t totalSize = image.total() * image.elemSize();
        unsigned char* alignedData = nullptr;

        // For POSIX systems
        if (posix_memalign(reinterpret_cast<void**>(&alignedData), 16, totalSize) != 0) {
            needToFree = false;
            return nullptr;
        }

        // Copy image data to aligned buffer
        std::memcpy(alignedData, image.data, totalSize);

        needToFree = true;
        return alignedData;
    }
}

uint32_t* allocateAlignedUint32Array(size_t N) {
    void* ptr = nullptr;
    size_t totalSize = N * sizeof(uint32_t);

    // For POSIX systems
    if (posix_memalign(&ptr, 16, totalSize) != 0) {
        // Allocation failed
        return nullptr;
    }

    return reinterpret_cast<uint32_t*>(ptr);
}

// Function to free the aligned memory
void freeAlignedMemory(void* ptr) {
    free(ptr);  // For posix_memalign
}